@@ -10,22 +10,45 @@ use JSON qw/ to_json decode_json /;
 use LWP::UserAgent;
 use Set::Scalar;
 use Text::Tradition::Analysis::Result;
-use Text::Tradition::Directory;
 use Text::Tradition::Stemma;
 use TryCatch;
 
 use vars qw/ @EXPORT_OK $VERSION /;
 @EXPORT_OK = qw/ run_analysis group_variants analyze_variant_location wit_stringify /;
-$VERSION = "1.0";
+$VERSION = "2.0.0";
 
 
-my $SOLVER_URL = 'http://byzantini.st/cgi-bin/graphcalc.cgi';
+my $DEFAULT_SOLVER_URL = 'http://perf.byzantini.st/cgi-bin/graphcalc.cgi';
 my $unsolved_problems = {};
 
 =head1 NAME
 
 Text::Tradition::Analysis - functions for stemma analysis of a tradition
 
+=head1 DESCRIPTION
+
+Text::Tradition is a library for representation and analysis of collated
+texts, particularly medieval ones.  Where the Collation is the central
+feature of a Tradition, it may also have one or more stemmata associated
+with it, and these stemmata may be analyzed. This package provides the
+following modules:
+
+=over 4
+
+=item * L<Text::Tradition::HasStemma> - a role that will be composed into
+Text::Tradition objects, providing the ability for Text::Tradition::Stemma
+objects to be associated with them.
+
+=item * L<Text::Tradition::Stemma> - an object class that represents stemma
+hypotheses, both rooted (with a single archetype) and unrooted (e.g.
+phylogenetic trees).
+
+=item * Text::Tradition::Analysis (this package). Provides functions for
+the analysis of a given stemma against the collation within a given
+Tradition.
+
+=back
+
 =head1 SYNOPSIS
 
   use Text::Tradition;
@@ -34,29 +57,10 @@ Text::Tradition::Analysis - functions for stemma analysis of a tradition
     'name' => 'this is a text',
     'input' => 'TEI',
     'file' => '/path/to/tei_parallel_seg_file.xml' );
-  $t->enable_stemmata;
   $t->add_stemma( 'dotfile' => $stemmafile );
 
   my $variant_data = run_analysis( $tradition );
-  # Recalculate rank $n treating all orthographic variants as equivalent
-  my $reanalyze = analyze_variant_location( $tradition, $n, 0, 'orthographic' );
     
-=head1 DESCRIPTION
-
-Text::Tradition is a library for representation and analysis of collated
-texts, particularly medieval ones.  Where the Collation is the central feature of
-a Tradition, it may also have one or more temmata associated with it, and these stemmata may be analyzed. This package provides the following modules:
-
-=over 4
-
-=item * L<Text::Tradition::HasStemma> - a role that can be composed into Text::Tradition objects, providing the ability for Text::Tradition::Stemma objects to be associated with them.
-
-=item * L<Text::Tradition::Stemma> - an object class that represents stemma hypotheses, both rooted (with a single archetype) and unrooted (e.g. phylogenetic trees).
-
-=item * Text::Tradition::Analysis (this package). Provides functions for the analysis of a given stemma against the collation within a given Tradition.
-
-=back
-
 =head1 SUBROUTINES
 
 =head2 run_analysis( $tradition, %opts )
@@ -87,7 +91,6 @@ my $datafile = 't/data/florilegium_tei_ps.xml';
 my $tradition = Text::Tradition->new( 'input' => 'TEI',
                                       'name' => 'test0',
                                       'file' => $datafile );
-$tradition->enable_stemmata;
 my $s = $tradition->add_stemma( 'dotfile' => 't/data/florilegium.dot' );
 is( ref( $s ), 'Text::Tradition::Stemma', "Added stemma to tradition" );
 
@@ -167,13 +170,18 @@ sub run_analysis {
 		$collapse->insert( $opts{'merge_types'} );
 	}
 	
-	# Make sure we have a lookup DB for graph calculation results; this will die
-	# if calcdir or calcdsn isn't passed.
+	# If we have specified a local lookup DB for graph calculation results,
+	# make sure it exists and connect to it.
 	my $dir;
-	if( exists $opts{'calcdir'} ) {
-		$dir = delete $opts{'calcdir'}
-	} elsif ( exists $opts{'calcdsn'} ) {
-		$dir = Text::Tradition::Directory->new( dsn => $opts{'calcdsn'} );
+	if ( exists $opts{'calcdsn'} ) {
+		eval { require Text::Tradition::Directory };
+		if( $@ ) {
+			throw( "Could not instantiate a directory for " . $opts{'calcdsn'}
+				. ": $@" );
+		}
+		$opts{'dir'} = Text::Tradition::Directory->new( dsn => $opts{'calcdsn'} );
+	} elsif( !exists $opts{'solver_url'} ) {
+		$opts{'solver_url'} = $DEFAULT_SOLVER_URL;
 	}
 
 	# Get the stemma	
@@ -221,7 +229,7 @@ sub run_analysis {
 	# Run the solver
 	my $answer;
 	try {
-		$answer = solve_variants( $dir, @groups );
+		$answer = solve_variants( \%opts, @groups );
 	} catch ( Text::Tradition::Error $e ) {
 		if( $e->message =~ /IDP/ ) {
 			# Something is wrong with the solver; make the variants table anyway
@@ -313,6 +321,7 @@ sub group_variants {
 	my $check_for_gaps = Set::Scalar->new();
 	my %moved_wits;
 	my $has_transposition;
+	my @transp_acgap;
 	foreach my $tablewit ( @{$table->{'alignment'}} ) {
 		my $rdg = $tablewit->{'tokens'}->[$rank-1];
 		my $wit = $tablewit->{'witness'};
@@ -333,16 +342,30 @@ sub group_variants {
 			# Otherwise, record it...
 			$readings_at_rank{$rdg->{'t'}->id} = $rdg->{'t'};
 			# ...and grab any transpositions, and their relations.
-			my @transp = grep { $_->rank != $rank } $rdg->{'t'}->related_readings();
+			my @transp = grep { $_->rank != $rank } _all_related( $rdg->{'t'} );
 			foreach my $trdg ( @transp ) {
 				next if exists $readings_at_rank{$trdg->id};
 				$has_transposition = 1;
 				my @affected_wits = _table_witnesses( 
-					$table, $trdg, $lacunose, $aclabel );
+					$table, $trdg->rank, $trdg, $lacunose, $aclabel );
 				next unless @affected_wits;
 				map { $moved_wits{$_} = 1 } @affected_wits;
-				$transposed->{$trdg->id} = 
-					[ _table_witnesses( $table, $rdg->{'t'}, $lacunose, $aclabel ) ];
+				my @thisloc_wits = _table_witnesses( $table, $rank, $rdg->{'t'}, 
+					$lacunose, $aclabel );
+				# Check to see if our affected wits have layers that do something
+				# wacky.
+				my %transploc_gaps;
+				map { $transploc_gaps{$_} = 1 } 
+					_table_witnesses( $table, $trdg->rank, undef, $lacunose, $aclabel );
+				foreach my $aw ( @affected_wits ) {
+					if( $transploc_gaps{$aw.$aclabel} ) {
+						push( @thisloc_wits, $aw.$aclabel );
+						push( @transp_acgap, $aw.$aclabel );
+					}
+				}
+				# Record which witnesses we should count as already analyzed when we 
+				# get to the transposed reading's own rank.
+				$transposed->{$trdg->id} = \@thisloc_wits;
 				$readings_at_rank{$trdg->id} = $trdg;
 			}
 		# ...or it is empty, ergo a gap.
@@ -350,6 +373,9 @@ sub group_variants {
 			_add_to_witlist( $wit, $check_for_gaps, $aclabel );
 		}
 	}
+	# Push all the transposition layer gaps onto our list
+	$check_for_gaps->insert( @transp_acgap );
+	# Now remove from our 'gaps' any witnesses known to have been dealt with elsewhere.
 	my $gap_wits = Set::Scalar->new();
 	map { _add_to_witlist( $_, $gap_wits, $aclabel ) 
 		unless $moved_wits{$_} } $check_for_gaps->members;
@@ -361,11 +387,11 @@ sub group_variants {
 		next if exists $grouped_readings->{$rdg->id} 
 			&& $grouped_readings->{$rdg->id} eq 'COLLAPSE';
 		# Get the witness list, including from readings collapsed into this one.
-		my @wits = _table_witnesses( $table, $rdg, $lacunose, $aclabel );
+		my @wits = _table_witnesses( $table, $rdg->rank, $rdg, $lacunose, $aclabel );
 		if( $collapse && $collapse->size ) {
 			my $filter = sub { $collapse->has( $_[0]->type ) };
 			foreach my $other ( $rdg->related_readings( $filter ) ) {
-				my @otherwits = _table_witnesses( $table, $other, $lacunose, $aclabel );
+				my @otherwits = _table_witnesses( $table, $other->rank, $other, $lacunose, $aclabel );
 				push( @wits, @otherwits );
 				$grouped_readings->{$other->id} = 'COLLAPSE';
 			}
@@ -394,19 +420,53 @@ sub group_variants {
 	return $grouped_readings;
 }
 
+sub _all_related {
+	# Except by repetition
+	my $rdg = shift;
+	my $c = $rdg->collation;
+	my @check = ( $rdg );
+	my %seen;
+	while( @check ) {
+		my @next;
+		foreach my $ck ( @check ) {
+			$seen{"$ck"} = 1;
+			push( @next, grep { !$seen{"$_"} } 
+				$ck->related_readings( sub { $_[0]->type ne 'repetition' } ) );
+		}
+		@check = @next;
+	}
+			
+		
+	my @all = map { $c->reading( $_ ) } keys %seen;
+	return @all;
+}
+	
+
 # Helper function to query the alignment table for all witnesses (a.c. included)
 # that have a given reading at its rank.
 sub _table_witnesses {
-	my( $table, $trdg, $lacunose, $aclabel ) = @_;
-	my $tableidx = $trdg->rank - 1;
+	my( $table, $rank, $trdg, $lacunose, $aclabel ) = @_;
+	my $tableidx = $rank - 1;
 	my $has_reading = Set::Scalar->new();
 	foreach my $row ( @{$table->{'alignment'}} ) {
 		my $wit = $row->{'witness'};
 		next if _is_lacunose( $wit, $lacunose, $aclabel );
 		my $rdg = $row->{'tokens'}->[$tableidx];
-		next unless exists $rdg->{'t'} && defined $rdg->{'t'};
-		_add_to_witlist( $wit, $has_reading, $aclabel )
-			if $rdg->{'t'}->id eq $trdg->id;
+		if( $trdg ) {
+			# We have some positive reading we want.
+			next unless exists $rdg->{'t'} && defined $rdg->{'t'};
+			if( $trdg->is_lacuna ) {
+				_add_to_witlist( $wit, $has_reading, $aclabel )
+				if $rdg->{'t'}->is_lacuna;
+			} else {
+				_add_to_witlist( $wit, $has_reading, $aclabel )
+					if $rdg->{'t'}->id eq $trdg->id;
+			}
+		} else {
+			# We want the omissions.
+			next if exists $rdg->{'t'} && defined $rdg->{'t'};
+			_add_to_witlist( $wit, $has_reading, $aclabel )
+		}
 	}
 	return $has_reading->members;
 }
@@ -454,8 +514,8 @@ sub _check_transposed_consistency {
 	if( @doubled == scalar keys %seen_wits ) {
 		foreach my $rdg ( keys %$groupings ) {
 			if( !$thisrank{$rdg} ) {
-				my $groupstr = wit_stringify( $groupings->{$rdg} );
-				my ( $matched ) = grep { $groupstr eq wit_stringify( $groupings->{$_} ) }
+				# Groupings are Set::Scalar objects so we can compare them outright.
+				my ( $matched ) = grep { $groupings->{$rdg} == $groupings->{$_} }
 					keys %thisrank;
 				delete $groupings->{$rdg};
 				# If we found a group match, assume there is a symmetry happening.
@@ -531,14 +591,10 @@ The answer has the form
 =cut
 
 sub solve_variants {
-	my( @groups ) = @_;
+	my( $opts, @groups ) = @_;
 	
-	# Are we using a local result directory, or did we pass an empty value
-	# for the directory?
-	my $dir;
-	unless( ref( $groups[0] ) eq 'HASH' ) {
-		$dir = shift @groups;
-	}
+	# Are we using a local result directory?
+	my $dir = $opts->{dir};
 
 	## For each graph/group combo, make a Text::Tradition::Analysis::Result
 	## object so that we can send it off for IDP lookup.
@@ -564,12 +620,13 @@ sub solve_variants {
 		my $scope = $dir->new_scope;
 		map { $results{$_} = $dir->lookup( $_ ) || $problems{$_} } keys %problems;
 	} else {
+		# print STDERR "Using solver at " . $opts->{solver_url} . "\n";
 		my $json = JSON->new->allow_blessed->convert_blessed->utf8->encode( 
 			[ values %problems ] );
 		# Send it off and get the result
 		# print STDERR "Sending request: " . decode_utf8( $json ) . "\n";
 		my $ua = LWP::UserAgent->new();
-		my $resp = $ua->post( $SOLVER_URL, 'Content-Type' => 'application/json', 
+		my $resp = $ua->post( $opts->{solver_url}, 'Content-Type' => 'application/json', 
 							  'Content' => $json );	
 		my $answer;	
 		if( $resp->is_success ) {
@@ -773,7 +830,17 @@ sub _resolve_parent_relationships {
 		my $prep = $pobj ? $pobj->id . ' (' . $pobj->text . ')' : $p;
 		my $phash = { 'label' => $prep };
 		if( $pobj ) {
+			# Get the attributes of the parent object while we are here
+			$phash->{'text'} = $pobj->text if $pobj;
+			if( $pobj && $pobj->does('Text::Tradition::Morphology') ) {
+				$phash->{'is_nonsense'} = $pobj->is_nonsense;
+				$phash->{'is_ungrammatical'} = $pobj->grammar_invalid;
+			}
+			# Now look at the relationship
 			my $rel = $c->get_relationship( $p, $rid );
+			if( $rel && $rel->type eq 'collated' ) {
+				$rel = undef;
+			}
 			if( $rel ) {
 				_add_to_hash( $rel, $phash );
 			} elsif( $rdg ) {
@@ -810,14 +877,15 @@ sub _resolve_parent_relationships {
 			} else {
 				$phash->{relation} = { type => 'deletion' };
 			}
-			# Get the attributes of the parent object while we are here
-			$phash->{'text'} = $pobj->text if $pobj;
-			if( $pobj && $pobj->does('Text::Tradition::Morphology') ) {
-				$phash->{'is_nonsense'} = $pobj->is_nonsense;
-				$phash->{'is_ungrammatical'} = $pobj->grammar_invalid;
-			}
 		} elsif( $p eq '(omitted)' ) {
-			$phash->{relation} = { type => 'addition' };
+			# Check to see if the reading in question is a repetition.
+			my @reps = $rdg->related_readings( 'repetition' );
+			if( @reps ) {
+				$phash->{relation} = { type => 'repetition', 
+					annotation => "of reading @reps" };
+			} else {
+				$phash->{relation} = { type => 'addition' };
+			}
 		}
 		# Save it
 		$rdgparents->{$p} = $phash;
@@ -830,6 +898,14 @@ sub _add_to_hash {
 	$phash->{relation}->{transposed} = 1 if $is_transposed;
 	$phash->{relation}->{annotation} = $rel->annotation
 		if $rel->has_annotation;
+	# Get all the relevant relationship info.
+	foreach my $prop ( qw/ non_independent is_significant / ) {
+		$phash->{relation}->{$prop} = $rel->$prop;
+	}
+	# Figure out if the variant was judged revertible.
+	my $is_a = $rel->reading_a eq $phash->{text};
+	$phash->{revertible} = $is_a 
+		? $rel->a_derivable_from_b : $rel->b_derivable_from_a;
 }
 
 =head2 similar( $word1, $word2 )
